from django.http import HttpResponse
from django.shortcuts import render

from markupy.tag import H1, Body, Html


def index(request):
    return HttpResponse(Html[Body[H1["Hi Django!"]]])


def template(request):
    context = {
        "content": H1["This is markupy title!"],
    }
    return render(request, "markupyapp/base.html", context)



from dataclasses import dataclass, field

import pytest

from markupy import Component, Fragment, View, tag
from markupy.exception import MarkupyError


class ComponentElement(Component):
    def __init__(self, id: str):
        self.id = id

    def render(self) -> View:
        return tag.Div(id=self.id)


def test_component_element() -> None:
    assert str(ComponentElement("component")) == """<div id="component"></div>"""


class ComponentFragment(Component):
    def render(self) -> View:
        return Fragment[tag.Div, tag.Img]


def test_uninitialized_component() -> None:
    with pytest.raises(MarkupyError):
        tag.P[ComponentFragment]


def test_component_fragment() -> None:
    assert str(ComponentFragment()) == """<div></div><img>"""


class ComponentInComponent(Component):
    def render(self) -> View:
        return Fragment[tag.Input, ComponentElement("inside")]


def test_component_in_component() -> None:
    assert str(ComponentInComponent()) == """<input><div id="inside"></div>"""


class ComponentAsComponent(Component):
    def render(self) -> View:
        return ComponentElement("other")


def test_component_as_component() -> None:
    assert str(ComponentAsComponent()) == """<div id="other"></div>"""


class ContentComponent(Component):
    def __init__(self, id: str) -> None:
        super().__init__()
        self.id = id

    def render(self) -> View:
        return tag.H1(".title.header", id=self.id)[self.render_content()]


def test_component_content() -> None:
    assert (
        str(ContentComponent(id="test")["Hello", tag.Div[tag.Input]])
        == """<h1 class="title header" id="test">Hello<div><input></div></h1>"""
    )


def test_component_content_escape() -> None:
    # Make sure component contents are not re-escaped when assigned to element children
    assert (
        str(ContentComponent(id="test")['He>"llo'])
        == """<h1 class="title header" id="test">He&gt;&#34;llo</h1>"""
    )


class TypeErrorComponent(Component):
    def render(self) -> str:  # type:ignore
        return "Hello"


def test_type_error_component() -> None:
    with pytest.raises(MarkupyError):
        str(TypeErrorComponent())


class SuperErrorComponent(Component):
    def __init__(self, *, id: str) -> None:
        # Missing the call to super().__init()
        # calls to __getitem__() will fail
        self.id = id

    def render(self) -> View:
        return tag.H1(id=self.id)


def test_super_error_component() -> None:
    assert str(SuperErrorComponent(id="foo")) == """<h1 id="foo"></h1>"""
    with pytest.raises(MarkupyError):
        SuperErrorComponent(id="foo")["bar"]


@dataclass
class DataComponent(Component):
    href: str = field(default="https://google.com")

    def render(self) -> View:
        return tag.A(href=self.href)[self.render_content()]


def test_dataclass_component() -> None:
    result = """<a href="https://google.com">Google</a>"""
    assert str(DataComponent()["Google"]) == result



from abc import abstractmethod
from collections.abc import Iterator
from typing import Any, final

from typing_extensions import Self

from ..exception import MarkupyError
from .fragment import Fragment
from .view import View


class Component(Fragment):
    __slots__ = ()

    def __init__(self) -> None:
        # Implementation here is useless but allows for a nice
        # argument-less super().__init__() autocomplete in user's IDE
        super().__init__()

    def __post_init__(self) -> None:
        # Allows for simple dataclass Component definition
        super().__init__()

    @final
    def __getitem__(self, content: Any) -> Self:
        if not hasattr(self, "_children"):
            raise MarkupyError(
                "Subclasses of <markupy.Component> must call `super().__init__()` if they override the default initializer."
            )

        return super().__getitem__(content)

    @abstractmethod
    def render(self) -> View: ...

    @final
    def render_content(self) -> View:
        # Use getattr here to avoid error in case super()__init__ hasn't been called
        if children := getattr(self, "_children", None):
            if len(children) == 1 and isinstance(children[0], View):
                # Only 1 view child: return it
                return children[0]
            else:
                # One non-view child or multiple children: wrap in a fragment
                # (do not use the [] syntax to avoid re-processing/re-escaping)
                fragment = Fragment()
                fragment._children = children
                return fragment

        else:
            # No children or super().__init__() hasn't been called and we are
            # missing the _children attribute: return empty view
            return View()

    @final
    def __iter__(self) -> Iterator[str]:
        node = self.render()
        if isinstance(node, View):  # type: ignore[unused-ignore]
            yield from node
        else:
            raise MarkupyError(
                f"`{type(self).__name__}.render()` must return an instance of <markupy.View> or one of its subclasses (Element, Fragment, Component)"
            )

    @final
    def __repr__(self) -> str:
        return f"<markupy.Component.{type(self).__name__}>"



from collections.abc import Iterable, Iterator
from inspect import isclass, isfunction, ismethod
from typing import Any, TypeAlias

from markupsafe import escape
from typing_extensions import Self

from ..exception import MarkupyError
from .view import View

ChildType: TypeAlias = str | View
ChildrenType: TypeAlias = tuple[ChildType, ...]


class Fragment(View):
    __slots__ = ("_children", "_safe")

    def __init__(self, *, safe: bool = False) -> None:
        super().__init__()
        self._children: ChildrenType = tuple()
        self._safe: bool = safe

    def _iter_node(self, node: Any) -> Iterator[ChildType]:
        if node is None or isinstance(node, bool):
            return
        elif isinstance(node, View):
            # View is Iterable, must check in priority
            yield node
        elif isinstance(node, Iterable) and not isinstance(node, str):
            for child in node:  # type: ignore[unused-ignore]
                yield from self._iter_node(child)
        elif isfunction(node) or ismethod(node) or isclass(node):
            # Allows to catch uncalled functions/methods or uninstanciated classes
            raise MarkupyError(
                f"Invalid child node {node!r} provided for {self!r}; Did you mean `{node.__name__}()` ?"
            )
        else:
            try:
                if s := str(node if self._safe else escape(node)):
                    yield s
            except Exception as e:
                raise MarkupyError(
                    f"Invalid child node {node!r} provided for {self!r}"
                ) from e

    # Use subscriptable [] syntax to assign children
    def __getitem__(self, content: Any) -> Self:
        if self._children:
            raise MarkupyError(f"Illegal attempt to redefine children of {self!r}")

        if children := tuple(c for c in self._iter_node(content)):
            instance = self._get_instance()
            instance._children = children
            return instance

        return self

    def __iter__(self) -> Iterator[str]:
        for node in self._children:
            if isinstance(node, View):
                yield from node
            else:
                yield node

    def _get_instance(self: Self) -> Self:
        return self

    def __repr__(self) -> str:
        return "<markupy.Fragment>"



import time

import django
from django.conf import settings
from django.template import Context
from django.template import Template as DjangoTemplate
from htpy import table, tbody, td, th, thead, tr
from jinja2 import Template as JinjaTemplate

from markupy.tag import Table, Tbody, Td, Th, Thead, Tr

settings.configure(
    TEMPLATES=[{"BACKEND": "django.template.backends.django.DjangoTemplates"}]
)
django.setup()

django_jinja_template = """
<table>
    <thead><tr><th>Row #</th></tr></thead>
    <tbody>
        {% for row in rows %}
        <tr class="row"><td data-value="{{ row }}">{{ row }}</td></tr>
        {% endfor %}
    </tbody>
</table>
"""

rows = list(range(50_000))


def render_markupy() -> str:
    return str(
        Table[
            Thead[Tr[Th["Row #"]]],
            Tbody[(Tr[Td[row]] for row in rows)],
        ]
    )


def render_markupy_attr() -> str:
    return str(
        Table[
            Thead[Tr[Th["Row #"]]],
            Tbody[
                (
                    Tr(".row")[
                        Td(f"#id-{row}.foo.bar", {"hello": "world"}, dataValue=row)[row]
                    ]
                    for row in rows
                )
            ],
        ]
    )


def render_htpy() -> str:
    return str(
        table[
            thead[tr[th["Row #"]]],
            tbody[(tr[td[row]] for row in rows)],
        ]
    )


def render_htpy_attr() -> str:
    return str(
        table[
            thead[tr[th["Row #"]]],
            tbody[
                (
                    tr(".row")[
                        td(f"#id-{row}.foo.bar", {"hello": "world"}, data_value=row)[
                            row
                        ]
                    ]
                    for row in rows
                )
            ],
        ]
    )


def render_django() -> str:
    return DjangoTemplate(django_jinja_template).render(Context({"rows": rows}))


def render_jinja() -> str:
    return JinjaTemplate(django_jinja_template).render(rows=rows)


tests = [
    render_markupy,
    render_markupy_attr,
    render_htpy,
    render_htpy_attr,
    render_django,
    render_jinja,
]

for func in tests:
    start = time.perf_counter()
    func()
    result = time.perf_counter() - start
    print(f"{func.__name__}: {result} seconds")



# type:ignore
import pytest
from flask import Flask, request, stream_with_context

from markupy import tag

app = Flask(__name__)


@app.route("/render")
def render():
    return str(tag.H1(".title")["render"])


@app.route("/stream")
def stream():
    return iter(tag.H1(".title")[request.args["name"]])


@app.route("/stream_context")
def stream_context():
    # Here stream_with_context is useless since the View is build before
    # the streaming starts and context is no longer needed
    return stream_with_context(tag.H1(".title")[request.args["name"]])


@pytest.fixture
def client():
    with app.test_client() as client:
        yield client


def test_render(client):
    response = client.get("/render")
    assert response.status_code == 200
    assert response.data.decode() == """<h1 class="title">render</h1>"""


def test_stream(client):
    response = client.get("/stream?name=stream")
    assert response.status_code == 200
    assert response.data.decode() == """<h1 class="title">stream</h1>"""


def test_stream_context(client):
    response = client.get("/stream_context?name=context")
    assert response.status_code == 200
    assert response.data.decode() == """<h1 class="title">context</h1>"""



# type: ignore
from django.conf import settings
from django.http import HttpResponse, StreamingHttpResponse
from django.urls import path

from markupy import tag

# --- Django minimal config ---
settings.configure(
    ROOT_URLCONF=__name__,
)


# --- The view ---
def render(request):
    return HttpResponse(tag.H1(".title")["render"])


def stream(request):
    return StreamingHttpResponse(iter(tag.H1(".title")["stream"]))


# --- URL config ---
urlpatterns = [
    path("render/", render),
    path("stream/", stream),
]


def test_render(client):
    response = client.get("/render/")
    assert response.status_code == 200
    assert response.content.decode() == """<h1 class="title">render</h1>"""


def test_stream(client):
    response = client.get("/stream/")
    assert response.status_code == 200
    assert (
        b"".join(response.streaming_content).decode()
        == """<h1 class="title">stream</h1>"""
    )



import cProfile

from markupy import View
from markupy.tag import Table, Tbody, Td, Th, Thead, Tr

rows = list(range(50_000))


def render() -> View:
    return Table[
        Thead[Tr[Th["Row #"]]],
        Tbody[(Tr(".row")[Td(dataValue=row)[row]] for row in rows)],
    ]


with cProfile.Profile() as pr:
    str(render())
    # Results can be wiewed with snakeviz
    # uvx snakeviz output.prof
    pr.dump_stats("output.prof")
    # pr.print_stats(sort="tottime")



import argparse
import sys

from .html import to_markupy


def main() -> None:
    parser = argparse.ArgumentParser(prog="html2markupy")

    parser.add_argument(
        "--selector",
        action=argparse.BooleanOptionalAction,
        help="Use the selector #id.class syntax instead of explicit `id` and `class_` attributes",
        default=True,
    )
    parser.add_argument(
        "--dict-attrs",
        action=argparse.BooleanOptionalAction,
        help="Prefer dict attributes",
        default=False,
    )
    parser.add_argument(
        "--tag-prefix",
        action=argparse.BooleanOptionalAction,
        help="Output mode for imports of markupy elements",
        default=False,
    )
    parser.add_argument(
        "input",
        type=argparse.FileType("r"),
        nargs="?",
        default=sys.stdin,
        help="input HTML from file or stdin",
    )

    args = parser.parse_args()
    try:
        html = args.input.read()
    except KeyboardInterrupt:
        print("\nInterrupted", file=sys.stderr)
        sys.exit(1)

    use_selector: bool = args.selector
    use_dict: bool = args.dicts_attrs
    use_import_tag: bool = args.tag_prefix

    print(
        to_markupy(
            html,
            use_selector=use_selector,
            use_dict=use_dict,
            use_import_tag=use_import_tag,
        )
    )



from collections.abc import Iterator, Mapping
from typing import Any, overload

from typing_extensions import Self, override

from ..exception import MarkupyError
from .attribute import AttributeDict, AttributeValue
from .shared import Shared


class Element(Shared):
    __slots__ = ("_attributes", "_name")

    def __init__(self, name: str, *, safe: bool = False, shared: bool = True) -> None:
        super().__init__(safe=safe, shared=shared)
        self._name = name
        self._attributes: str | None = None

    def __copy__(self) -> Self:
        return type(self)(self.name, shared=False)

    @property
    def name(self) -> str:
        return self._name

    def _tag_opening(self) -> str:
        if attributes := self._attributes:
            return f"<{self._name} {attributes}>"
        return f"<{self._name}>"

    def _tag_closing(self) -> str:
        return f"</{self._name}>"

    def __iter__(self) -> Iterator[str]:
        yield self._tag_opening()
        yield from super().__iter__()
        yield self._tag_closing()

    def __repr__(self) -> str:
        return f"<markupy.{type(self).__name__}.{self._name}>"

    # Use call syntax () to define attributes
    @overload
    def __call__(
        self,
        selector: str,
        attributes: Mapping[str, AttributeValue],
        **kwargs: AttributeValue,
    ) -> Self: ...
    @overload
    def __call__(self, selector: str, **kwargs: AttributeValue) -> Self: ...
    @overload
    def __call__(
        self, attributes: Mapping[str, AttributeValue], **kwargs: AttributeValue
    ) -> Self: ...
    @overload
    def __call__(self, **kwargs: AttributeValue) -> Self: ...
    def __call__(self, *args: Any, **kwargs: Any) -> Self:
        if self._attributes is not None:
            raise MarkupyError(
                f"Illegal attempt to redefine attributes for element {self!r}"
            )

        if self._children:
            raise MarkupyError(
                f"Illegal attempt to define attributes after children for element {self!r}"
            )

        selector: str | None = None
        attributes_dict: Mapping[str, AttributeValue] | None = None
        attributes_kwargs: Mapping[str, AttributeValue] = kwargs
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                # element(".foo")
                selector = arg
            elif isinstance(arg, Mapping):
                # element({"foo": "bar"})
                attributes_dict = arg  # type:ignore[unused-ignore]
            else:
                raise MarkupyError(
                    f"Invalid argument type {arg!r} for element {self!r}, expected `str` or `Mapping`"
                )
        elif len(args) == 2:
            # element(".foo", {"bar": "baz"})
            if not isinstance(args[0], str):
                raise MarkupyError(
                    f"Invalid first argument type {args[0]!r} for element {self!r}, expected `str`"
                )
            if not isinstance(args[1], Mapping):
                raise MarkupyError(
                    f"Invalid second argument type {args[1]!r} for element {self!r}, expected `Mapping`"
                )
            selector, attributes_dict = args
        elif len(args) > 2:
            raise MarkupyError(
                f"Invalid number of arguments provided for element {self!r}"
            )

        if not selector and not attributes_dict and not attributes_kwargs:
            return self

        attrs = AttributeDict()
        try:
            attrs.add_selector(selector)
        except Exception as e:
            raise MarkupyError(
                f"Invalid selector string `{selector}` for element {self!r}"
            ) from e
        try:
            attrs.add_dict(attributes_dict)
        except Exception as e:
            raise MarkupyError(
                f"Invalid dict attributes `{attributes_dict}` for element {self!r}"
            ) from e
        try:
            attrs.add_dict(attributes_kwargs, rewrite_keys=True)
        except Exception as e:
            raise MarkupyError(
                f"Invalid keyword attributes `{attributes_kwargs}` for element {self!r}"
            ) from e

        if attributes := str(attrs):
            el = self._get_instance()
            el._attributes = attributes
            return el

        return self


class HtmlElement(Element):
    __slots__ = ()

    @override
    def __iter__(self) -> Iterator[str]:
        yield "<!doctype html>"
        yield from super().__iter__()


class VoidElement(Element):
    __slots__ = ()

    @override
    def __iter__(self) -> Iterator[str]:
        yield self._tag_opening()

    @override
    def __getitem__(self, children: Any) -> Self:
        raise MarkupyError(f"Void element {self!r} cannot contain children")


class CommentElement(Element):
    __slots__ = ()

    @override
    def _tag_opening(self) -> str:
        return "<!--"

    @override
    def _tag_closing(self) -> str:
        return "-->"

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Self:
        raise MarkupyError(f"Comment element {self!r} cannot have attributes")


class SafeElement(Element):
    __slots__ = ()

    def __init__(self, name: str, *, shared: bool = True) -> None:
        super().__init__(name, safe=True, shared=shared)



from ._private.component import Component
from ._private.html import to_markupy as html2markupy
from ._private.shared import Shared
from ._private.view import View

Fragment = Shared()

__all__ = ["Component", "Fragment", "View", "html2markupy"]
__version__ = "1.4.0"



import typing as t
from datetime import date

import pytest
from markupsafe import Markup

from markupy.exception import MarkupyError
from markupy.tag import Button, Div, Input, Th, _


def test_attribute() -> None:
    assert str(Input(attr="disabled")) == '<input attr="disabled">'
    assert str(Input(disabled="disabled")) == "<input disabled>"


def test_true_value() -> None:
    assert str(Input(disabled=True)) == "<input disabled>"
    assert str(Input(disabled=1)) == "<input disabled>"
    assert str(Input(attr=True)) == "<input attr>"
    assert str(Input(attr=1)) == '<input attr="1">'


def test_false_value() -> None:
    assert str(Input(disabled=False)) == "<input>"
    assert str(Input(disabled=0)) == "<input>"
    assert str(Input(attr=False)) == "<input>"
    assert str(Input(attr=0)) == '<input attr="0">'


def test_none_value() -> None:
    assert str(Input(disabled=None)) == "<input>"
    assert str(Input(attr=None)) == "<input>"


def test_empty_value() -> None:
    # Different behaviour for boolean attributes vs regular
    assert str(Input(disabled="")) == """<input disabled>"""
    assert str(Input(attr="")) == """<input attr="">"""
    assert str(Input(id="", class_="")) == """<input>"""


def test_comment() -> None:
    with pytest.raises(MarkupyError):
        _(attr="foo")


class Test_class_names:
    def test_str(self) -> None:
        result = Div(class_='">foo bar')
        assert str(result) == '<div class="&#34;&gt;foo bar"></div>'

    def test_safestring(self) -> None:
        result = Div(class_=Markup('">foo bar'))
        assert str(result) == '<div class="&#34;&gt;foo bar"></div>'

    def test_list(self) -> None:
        result = Div(class_=['">foo', Markup('">bar'), None, "", "baz"])
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    def test_tuple(self) -> None:
        result = Div(class_=('">foo', Markup('">bar'), None, "", "baz"))
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    def test_dict(self) -> None:
        result = Div(
            class_={'">foo': True, Markup('">bar'): True, "x": False, "baz": True}
        )
        assert str(result) == '<div class="&#34;&gt;foo &#34;&gt;bar baz"></div>'

    def test_nested_dict(self) -> None:
        result = Div(
            class_=[
                '">list-foo',
                Markup('">list-bar'),
                {'">dict-foo': True, Markup('">list-bar'): True, "x": False},
            ]
        )
        assert (
            str(result)
            == """<div class="&#34;&gt;list-foo &#34;&gt;list-bar &#34;&gt;dict-foo &#34;&gt;list-bar"></div>"""
        )

    def test_false(self) -> None:
        result = str(Div(class_=False))
        assert result == "<div></div>"

    def test_none(self) -> None:
        result = str(Div(class_=None))
        assert result == "<div></div>"

    def test_no_classes(self) -> None:
        result = str(Div(class_={"foo": False}))
        assert result == "<div></div>"

    def test_selector_attr_mixed(self) -> None:
        result = str(Div(".foo", class_={"bar": True, "baz": False}))
        assert result == """<div class="foo bar"></div>"""


def test_dict_attributes() -> None:
    result = Div({"@click": 'hi = "hello"'})

    assert str(result) == """<div @click="hi = &#34;hello&#34;"></div>"""


def test_underscore() -> None:
    # Hyperscript (https://hyperscript.org/) uses _, make sure it works good.
    result = Div(_="foo")
    assert str(result) == """<div _="foo"></div>"""


def test_dict_attributes_avoid_replace() -> None:
    result = Div({"class_": "foo", "hello_hi": "abc"})
    assert str(result) == """<div class_="foo" hello_hi="abc"></div>"""


def test_dict_attribute_false() -> None:
    result = Div({"bool-false": False})
    assert str(result) == "<div></div>"


def test_dict_attribute_true() -> None:
    result = Div({"bool-true": True})
    assert str(result) == "<div bool-true></div>"


def test_uppercase_replacement() -> None:
    result = Button(hxPost="/foo")["click me!"]
    assert str(result) == """<button hx-post="/foo">click me!</button>"""


def test_at_replacement() -> None:
    result = Input(_input_debounce_500ms="fetchResults")
    assert str(result) == """<input @input.debounce.500ms="fetchResults">"""


def test_colon_replacement() -> None:
    result = Button(
        hxOn__htmx__configRequest="event.detail.parameters.example = 'Hello Scripting!'"
    )["Post Me!"]
    assert (
        str(result)
        == """<button hx-on:htmx:config-request="event.detail.parameters.example = &#39;Hello Scripting!&#39;">Post Me!</button>"""
    )


class Test_value_escape:
    pytestmark = pytest.mark.parametrize(
        "value",
        [
            '<"foo',
            Markup('<"foo'),
        ],
    )

    def test_dict(self, value: str) -> None:
        result = Div({"bar": value})
        assert str(result) == """<div bar="&lt;&#34;foo"></div>"""

    def test_kwarg(self, value: str) -> None:
        result = Div(**{"bar": value})
        assert str(result) == """<div bar="&lt;&#34;foo"></div>"""


def test_boolean_attribute_true() -> None:
    result = Button(disabled=True)
    assert str(result) == "<button disabled></button>"


def test_kwarg_attribute_none() -> None:
    result = Div(foo=None)
    assert str(result) == "<div></div>"


def test_dict_attribute_none() -> None:
    result = Div({"foo": None})
    assert str(result) == "<div></div>"


def test_boolean_attribute_false() -> None:
    result = Button(disabled=False)
    assert str(result) == "<button></button>"


def test_integer_attribute() -> None:
    result = Th(colspan=123, tabindex=0)
    assert str(result) == '<th colspan="123" tabindex="0"></th>'


def test_selector() -> None:
    result = Div("#myid.cls1.cls2")

    assert str(result) == """<div id="myid" class="cls1 cls2"></div>"""


def test_selector_only_id() -> None:
    result = Div("#myid")
    assert str(result) == """<div id="myid"></div>"""


def test_selector_only_classes() -> None:
    result = Div(".foo.bar")
    assert str(result) == """<div class="foo bar"></div>"""


def test_selector_empty_classes() -> None:
    result = Div(".foo..bar.")
    assert str(result) == """<div class="foo bar"></div>"""


def test_selector_classes_space_separator() -> None:
    result = Div("foo bar")
    assert str(result) == """<div class="foo bar"></div>"""


def test_selector_bad_type() -> None:
    with pytest.raises(MarkupyError):
        Div({"oops": "yes"}, {})  # type: ignore


def test_invalid_number_of_attributes() -> None:
    with pytest.raises(MarkupyError):
        Div("#id.cls", {"attr": "val"}, "other")  # type: ignore


def test_id_class_and_kwargs() -> None:
    result = Div("#theid", for_="hello", dataFoo="<bar")
    assert str(result) == """<div id="theid" for="hello" data-foo="&lt;bar"></div>"""


def test_attrs_and_kwargs() -> None:
    result = Div({"a": "1", "for": "a"}, for_="b", b="2")
    assert str(result) == """<div a="1" for="b" b="2"></div>"""


def test_class_priority() -> None:
    result = Div(".a", {"class": "b"}, class_="c")
    assert str(result) == """<div class="a b c"></div>"""

    result = Div(".a", {"class": "b"})
    assert str(result) == """<div class="a b"></div>"""


def test_attribute_priority() -> None:
    result = Div({"foo": "a"}, foo="b")
    assert str(result) == """<div foo="b"></div>"""


@pytest.mark.parametrize("not_an_attr", [1234, b"foo", object(), object, 1, 0, None])
def test_invalid_attribute_key(not_an_attr: t.Any) -> None:
    with pytest.raises(MarkupyError):
        str(Div({not_an_attr: "foo"}))


@pytest.mark.parametrize(
    "not_an_attr",
    [12.34, b"foo", object(), object],
)
def test_invalid_attribute_value(not_an_attr: t.Any) -> None:
    with pytest.raises(MarkupyError):
        Div(foo=not_an_attr)


def test_attribute_redefinition() -> None:
    with pytest.raises(MarkupyError):
        Div(id="hello")(class_="world")


@pytest.mark.parametrize("selector", ["", "   ", "  #  ", "  .  "])
def test_empty_selector(selector: str) -> None:
    result = Div(selector)
    assert str(result) == """<div></div>"""


def test_selector_strip() -> None:
    result = Div(" #myid .myclass .other ")
    assert str(result) == """<div id="myid" class="myclass other"></div>"""


@pytest.mark.parametrize(
    "key",
    ["", "  foo  ", "foo bar", '<"foo', Markup('<"foo'), date.today()],
)
def test_invalid_key(key: str) -> None:
    with pytest.raises(MarkupyError):
        Div({key: "bar"})
    with pytest.raises((MarkupyError, TypeError)):
        Div(**{key: "bar"})


def test_attribute_case() -> None:
    result = Div({"BAR": "foo", "bAr": "hello"}, bar="baz")
    # If not properly managed, could become <div BAR="foo" bar="baz"></div>
    assert str(result) == """<div bar="baz"></div>"""


def test_selector_invalid_id_position() -> None:
    with pytest.raises(MarkupyError):
        Div(".bar#foo")


def test_selector_multiple_id() -> None:
    with pytest.raises(MarkupyError):
        Div("#foo#bar")


def test_selector_empty_id() -> None:
    assert str(Div("# foo bar")) == """<div class="foo bar"></div>"""
    assert str(Div("#.foo.bar")) == """<div class="foo bar"></div>"""



from copy import copy
from typing import final

from typing_extensions import Self

from .fragment import Fragment


class Shared(Fragment):
    __slots__ = ("_shared",)

    def __init__(self, *, safe: bool = False, shared: bool = True) -> None:
        super().__init__(safe=safe)
        self._shared: bool = shared

    def __copy__(self) -> Self:
        return type(self)(shared=False)

    def __call__(self) -> Self:
        return self

    @final
    def _get_instance(self: Self) -> Self:
        # When imported, elements are loaded from a shared instance
        # Make sure we re-instantiate them on setting attributes/children
        # to avoid sharing attributes/children between multiple instances
        if self._shared:
            return copy(self)
        return self

    # Avoid having Django "call" a markupy fragment (or element) that is injected into a template.
    # Setting do_not_call_in_templates will prevent Django from doing an extra call:
    # https://docs.djangoproject.com/en/5.0/ref/templates/api/#variables-and-lookups
    do_not_call_in_templates = True



from collections.abc import Iterable, Iterator, Mapping, Sequence
from functools import lru_cache
from re import match as re_match
from re import sub as re_sub
from typing import TypeAlias

from markupsafe import escape

from ..exception import MarkupyError

ClassNamesDict: TypeAlias = Mapping[str, bool | None]
ClassNamesSequence: TypeAlias = Sequence[None | str | ClassNamesDict]
ClassAttributeValue: TypeAlias = ClassNamesSequence | ClassNamesDict
PrimitiveAttributeValue: TypeAlias = None | bool | str | int
AttributeValue: TypeAlias = PrimitiveAttributeValue | ClassAttributeValue

# https://html.spec.whatwg.org/multipage/indices.html#attributes-3
BOOLEAN_ATTRIBUTES: set[str] = {
    "allowfullscreen",
    "async",
    "autofocus",
    "autoplay",
    "checked",
    "controls",
    "default",
    "defer",
    "disabled",
    "formnovalidate",
    "inert",
    "ismap",
    "itemscope",
    "loop",
    "multiple",
    "muted",
    "nomodule",
    "novalidate",
    "open",
    "playsinline",
    "readonly",
    "required",
    "reversed",
    "selected",
}


def _classes_to_str(classes: Iterable[str]) -> str:
    return " ".join(map(lambda c: c.strip(), filter(None, classes)))


def _iter_classes_dict(dct: ClassNamesDict) -> Iterator[str]:
    for k, v in dct.items():
        if v:
            yield k


def _iter_classes_seq(seq: ClassNamesSequence) -> Iterator[str]:
    for v in seq:
        if not v:
            continue
        elif isinstance(v, Mapping):
            yield from _iter_classes_dict(v)
        else:
            yield v


@lru_cache(maxsize=1000)
def _rewrite_attr_key(key: str) -> str:
    # Leading underscore "_" not followed by another "_" -> "@"
    key = re_sub(r"^(?:_)([^_]+)", r"@\1", key)
    # Trailing underscore "_" is meaningless and is used to escape protected
    # keywords that might be used as attr keys such as class_ and for_
    key = key.removesuffix("_")
    # Upper case -> "-"
    key = "-".join(filter(None, re_sub(r"([A-Z])", r" \1", key).split())).lower()
    # Double underscore -> ":"
    key = key.replace("__", ":")
    # Single underscore -> "."
    key = key.replace("_", ".")
    return key


def is_boolean_attribute(name: str) -> bool:
    return name in BOOLEAN_ATTRIBUTES


def _format_key_value(key: str, value: AttributeValue) -> str:
    if value is True:
        return key
    return f'{key}="{escape(str(value))}"'


class AttributeDict(dict[str, AttributeValue]):
    __slots__ = ()

    def __setitem__(self, key: str, value: AttributeValue) -> None:
        if value is None or value is False:
            # Discard False and None valued attributes for all attributes
            return

        key = key.lower()

        if value is True:
            pass
        elif is_boolean_attribute(key):
            if value == 0:
                return
            value = True
        elif value == "" and key in ("id", "class"):
            # Discard empty id or class attributes
            return

        if key == "class" and self.get(key):
            return super().__setitem__(key, f"{self['class']} {value}")
        else:
            return super().__setitem__(key, value)

    def __str__(self) -> str:
        return " ".join(_format_key_value(k, v) for k, v in self.items())

    def add_selector(self, selector: str | None) -> None:
        if not selector:
            # Empty selector or None
            return

        selector = selector.replace(".", " ").strip()
        parts = selector.split()
        hash_indexes = [i for i, c in enumerate(selector) if c == "#"]

        if len(hash_indexes) > 1:
            raise MarkupyError("Id must be defined only once in selector")

        elif len(hash_indexes) == 1:
            if hash_indexes[0] != 0:
                raise MarkupyError("Id must be defined at the start of selector")

            self["id"] = parts[0][1:]
            self["class"] = _classes_to_str(parts[1:])

        else:
            self["class"] = _classes_to_str(parts)

    def add_dict(
        self,
        dct: Mapping[str, AttributeValue] | None,
        *,
        rewrite_keys: bool = False,
    ) -> None:
        if not dct:
            # Empty dict or None
            return
        for key, value in dct.items():
            if rewrite_keys:
                if not key.isidentifier():
                    # Might happen when using the **{} syntax
                    raise MarkupyError(f"Attribute `{key}` has invalid name")
                elif key != "_":
                    # Preserve single _ for hyperscript
                    key = _rewrite_attr_key(key)
            else:
                # Coming from dict arg, need to secure user input
                if not isinstance(key, str):  # pyright: ignore [reportUnnecessaryIsInstance]
                    raise MarkupyError(f"Attribute {key!r} must be a string")
                if escape(str(key)) != key or not re_match(r"^\S+$", key):
                    raise MarkupyError(f"Attribute `{key}` has invalid name")

            if isinstance(value, PrimitiveAttributeValue):
                self[key] = value
                continue

            elif key == "class":
                if isinstance(value, Mapping):
                    classes = _iter_classes_dict(value)
                else:
                    classes = _iter_classes_seq(value)
                self[key] = _classes_to_str(classes)
                continue

            raise MarkupyError(f"Invalid value type {value!r} for attribute `{key}`")



from collections.abc import Iterator
from typing import final

from markupsafe import Markup


class View:
    __slots__ = ()

    def __init__(self) -> None:
        super().__init__()

    def __iter__(self) -> Iterator[str]:
        yield from ()

    @final
    def __str__(self) -> str:
        # Return needs to be Markup and not plain str
        # to be properly injected in template engines
        return Markup("".join(self))

    def __repr__(self) -> str:
        return "<markupy.View>"

    # Allow starlette Response.render to directly render this element without
    # explicitly casting to str:
    # https://github.com/encode/starlette/blob/5ed55c441126687106109a3f5e051176f88cd3e6/starlette/responses.py#L44-L49
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes:
        return str(self).encode(encoding, errors)

    __html__ = __str__



from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.routing import Route

from markupy.tag import H1, Body, Html


async def index(request: Request) -> HTMLResponse:
    return HTMLResponse(Html[Body[H1["Hi Starlette!"]]])


# Run it with `uv run uvicorn html_response:app``
app = Starlette(
    routes=[Route("/", index)],
)



# type: ignore
from starlette.responses import HTMLResponse, StreamingResponse
from starlette.testclient import TestClient

from markupy import tag


async def render(scope, receive, send):
    assert scope["type"] == "http"
    response = HTMLResponse(tag.H1(".title")["render"])
    await response(scope, receive, send)


async def stream(scope, receive, send):
    assert scope["type"] == "http"
    response = StreamingResponse(iter(tag.H1(".title")["stream"]))
    await response(scope, receive, send)


def test_render() -> None:
    client = TestClient(render)
    response = client.get("/")
    assert response.text == """<h1 class="title">render</h1>"""


def test_stream() -> None:
    client = TestClient(stream)
    response = client.get("/")
    assert response.text == """<h1 class="title">stream</h1>"""



from typing import Iterator

import pytest

from markupy import Fragment
from markupy.exception import MarkupyError
from markupy.tag import Div, I, P, Tr


def test_render_direct() -> None:
    assert str(Fragment["Hello ", None, I["World"]]) == """Hello <i>World</i>"""


def test_render_as_child() -> None:
    assert (
        str(P["Say: ", Fragment["Hello ", None, I["World"]], "!"])
        == """<p>Say: Hello <i>World</i>!</p>"""
    )


def test_render_nested() -> None:
    assert str(Fragment[Fragment["Hel", "lo "], "World"]) == """Hello World"""


def test_render_embedded() -> None:
    assert (
        str(P[Fragment["Good ", I["morning"]], " ", I["World"]])
        == """<p>Good <i>morning</i> <i>World</i></p>"""
    )


def test_safe() -> None:
    assert str(Fragment['>"']) == """&gt;&#34;"""


def test_iter() -> None:
    assert list(Fragment["Hello ", None, I["World"]]) == [
        "Hello ",
        "<i>",
        "World",
        "</i>",
    ]


def test_element() -> None:
    result = Fragment[Div["a"]]
    assert str(result) == "<div>a</div>"


def test_multiple_element() -> None:
    result = Fragment[Tr["a"], Tr["b"]]
    assert str(result) == "<tr>a</tr><tr>b</tr>"


def test_list() -> None:
    result = Fragment[[Tr["a"], Tr["b"]]]
    assert str(result) == "<tr>a</tr><tr>b</tr>"


def test_none() -> None:
    result = Fragment[None]
    assert str(result) == ""


def test_string() -> None:
    result = Fragment["hello!"]
    assert str(result) == "hello!"


def test_class() -> None:
    class Test:
        def __call__(self) -> str:
            return self.message()

        def __str__(self) -> str:
            return self()

        def message(self) -> str:
            return "hello"

    assert str(Fragment[Test()]) == "hello"
    assert str(Fragment[Test().message()]) == "hello"
    with pytest.raises(MarkupyError):
        Fragment[Test]
    with pytest.raises(MarkupyError):
        Fragment[Test().message]


def test_lambda() -> None:
    with pytest.raises(MarkupyError):
        Fragment[lambda: "hello"]


def test_function() -> None:
    def test() -> str:
        return "hello"

    assert str(Fragment[test()]) == "hello"
    with pytest.raises(MarkupyError):
        Fragment[test]


def test_generator() -> None:
    def generator() -> Iterator[str]:
        yield "hello"
        yield "world"

    assert str(Fragment[generator()]) == "helloworld"
    with pytest.raises(MarkupyError):
        Fragment[generator]



import typing

from django.template import Context, Engine  # type:ignore
from jinja2 import Environment

from markupy import tag


@typing.no_type_check
def test_django() -> None:
    template = Engine().from_string("<div>{{ content }}</div>")
    context = Context({"content": tag.H1(".title")["django"]})
    rendered = template.render(context)
    assert rendered == """<div><h1 class="title">django</h1></div>"""


def test_jinja2() -> None:
    env = Environment(autoescape=True)
    template = env.from_string("<div>{{ content }}</div>")
    rendered = template.render(content=tag.H1(".title")["jinja"])
    assert rendered == """<div><h1 class="title">jinja</h1></div>"""
